library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity fdc_video is
  port (
    reset : in std_logic;
    clk   : in std_logic; -- 50 MHz

    read       : in  std_logic;
    write      : in  std_logic;
    chipselect : in  std_logic;
    address    : in  unsigned(4 downto 0);
    readdata   : out unsigned(15 downto 0);
    writedata  : in  unsigned(15 downto 0);

    VGA_CLK,                         -- Clock
    VGA_HS,                          -- H_SYNC
    VGA_VS,                          -- V_SYNC
    VGA_BLANK,                       -- BLANK
    VGA_SYNC : out std_logic;        -- SYNC
    VGA_R,                           -- Red[9:0]
    VGA_G,                           -- Green[9:0]
    VGA_B : out std_logic_vector(9 downto 0); -- Blue[9:0]

    LEDR : out std_logic_vector(17 downto 0);
    LEDG : out std_logic_vector(8 downto 0);

    GPIO_0 : in std_logic_vector(35 downto 0);

    -- Control Singals for SRAM
    SRAM_DQ   : inout std_logic_vector(15 downto 0);
    SRAM_ADDR : out std_logic_vector(17 downto 0);
    SRAM_UB_N,
    SRAM_LB_N,
    SRAM_WE_N,
    SRAM_CE_N,
    SRAM_OE_N : out std_logic
  );
end fdc_video;

architecture rtl of fdc_video is

COMPONENT TilesRam PORT (
    clk : in std_logic;
    we : in std_logic;
    a : in unsigned(14 downto 0);
    di : in unsigned(15 downto 0);
    do : out unsigned(15 downto 0)
  );
END COMPONENT;

COMPONENT counter
  port(
    Clk, Reset: in std_logic;
    Q: out integer
  );
end COMPONENT;
  -- Video parameters

  constant HTOTAL       : integer := 800;
  constant HSYNC        : integer := 96;
  constant HBACK_PORCH  : integer := 48;
  constant HACTIVE      : integer := 640;
  constant HFRONT_PORCH : integer := 16;

  constant VTOTAL       : integer := 525;
  constant VSYNC        : integer := 2;
  constant VBACK_PORCH  : integer := 33;
  constant VACTIVE      : integer := 480;
  constant VFRONT_PORCH : integer := 10;

  -- Opcodes

  -- Write a 16-bit R5G6B5 pixel to where the pixel counter is pointing in SRAM
  constant OP_WRITE : unsigned := "00000";

  -- Which part of the 1024-row space should we start showing?
  constant OP_SET_VERT_OFFSET : unsigned := "00001";

  -- Change pixel counter to the specified row in the 1024-row space
  -- Should also sets the column-counter to 0
  constant OP_SEEK_ROW : unsigned := "00100";

  -- Which of the 64 tiles is the one we will be writing to next?
  constant OP_TILE_SET_CURRENT : unsigned := "00101";

  -- Commit the 16-bit writedata as four 4-bit pixels to the current area in the tile
  -- pointer
  constant OP_TILE_WRITE : unsigned := "00110";

  -- Move the tile pointer to this value. Tiles are stored in raster-order and the pointer
  -- counts four 4-bit pixels at a time (e.g. 16 bit words)
  constant OP_TILE_SEEK : unsigned := "00111";

  -- Selects which tile the subsequent OP_TILE_CHANGE_TO command will change.
  constant OP_TILE_TO_CHANGE : unsigned := "01000";
  constant OP_TILE_CHANGE_TO : unsigned := "01010";

  -- Usually, we want to set this to 2 to set off the animation to next tile
  constant OP_TILE_SET_ANIM_STATE : unsigned := "01011";

  -- Set whether the tile in the tilematrix that we're currently setting should be tinted
  constant OP_TINT : unsigned := "01100";

  -- Set what tile will come up next as the numbers flip
  constant OP_TILE_TRANSFORMER_CHANGE: unsigned:="01101";

  -- Signals for the video controller
  signal Hcount : unsigned(9 downto 0) := (others => '0'); -- Horizontal position (0-800)
  signal Vcount : unsigned(9 downto 0) := (others => '0'); -- Vertical position (0-524)
  signal EndOfLine : std_logic := '0';
  signal EndOfField : std_logic := '0';

  type animationInfo is array(0 to 207) of integer range 0 to 32; -- animation matrix
  constant animation : animationInfo := (
    32,0,1,2,3,4,5,6,8,9,10,11,12,13,14,15, -- 1st step of animation
    32,32,0,1,2,3,4,6,7,8,9,11,12,13,14,15, -- 2nd step of animation... etc.
    32,32,32,0,1,2,3,5,6,7,9,10,11,13,14,15,
    32,32,32,32,32,0,1,3,4,6,7,9,10,12,13,15, -- each row has 16 values, one for each
    32,32,32,32,32,32,0,1,2,4,6,8,10,12,14,15, -- row of the tile. only 1 half done 
    32,32,32,32,32,32,32,0,1,2,4,6,10,12,14,15, -- at a time
    32,32,32,32,32,32,32,32,32,0,2,6,8,12,14,15,
    32,32,32,32,32,32,32,32,32,32,1,5,8,11,13,15,
    32,32,32,32,32,32,32,32,32,32,32,32,3,6,9,12,
    32,32,32,32,32,32,32,32,32,32,32,32,32,32,4,9,
    21,26,32,32,32,32,32,32,32,32,32,32,32,32,32,32, -- start of second half of animation
    19,22,26,29,32,32,32,32,32,32,32,32,32,32,32,32, -- (lower half only)
    19,21,23,24,25,27,29,32,32,32,32,32,32,32,32,32
  );

  type animationBigInfo is array (0 to 415) of integer range 0 to 32;
  constant animationBIG: animationBigInfo := (
    32,32,0,1,2,3,4,5,6,7,8,9,11,12,13,14,15,16,17,18,19,20,22,23,24,25,26,27,28,29,30,31,
    32,32,32,32,0,1,2,3,4,6,7,8,9,10,11,13,14,15,16,17,19,20,21,22,23,24,26,27,28,29,30,31,
    32,32,32,32,32,32,32,0,1,2,4,5,6,8,9,10,12,13,14,16,17,18,20,21,22,24,25,26,28,29,30,31,
    32,32,32,32,32,32,32,32,32,0,1,3,4,6,7,9,10,11,13,14,16,17,19,20,22,23,24,26,27,29,30,
    31,  -- Sorry for the wrapping...these lines were too long and messed up the report
    32,32,32,32,32,32,32,32,32,32,32,32,0,2,3,5,7,8,10,12,13,15,17,18,20,22,23,25,27,28,30,
    31,
    32,32,32,32,32,32,32,32,32,32,32,32,32,32,0,2,4,6,7,9,11,13,15,17,19,21,23,24,26,28,30,
    31,
    32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,0,2,5,7,9,12,14,16,18,21,23,25,28,30,
    31,
    32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,0,3,6,9,12,15,18,21,24,27,
    30,31,
    32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,1,6,10,15,20,24,
    29,31,
    32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,3,11,19,
    27,31,
    7,15,23,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,
    32,32,
    2,5,9,12,15,19,22,25,28,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,
    32,32,
    1,3,5,7,9,11,13,14,15,17,19,21,23,25,27,29,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,
    32
  );

  signal vga_hblank : std_logic := '0';
  signal vga_hsync : std_logic := '0';
  signal vga_vblank : std_logic := '0';
  signal vga_vsync : std_logic := '0';  -- Sync. signals

  signal clk25: std_logic := '0';
  signal incanvas: std_logic := '0';
  signal read_pointer: unsigned(19 downto 0) := (others => '0');
  signal write_pointer: unsigned(17 downto 0) := (others => '0');

  -- Ring-buffer type queue
  -- NOTE: if you ever see corruption on the right side of the screen
  -- a possibility is that the queue overflowed so increase size of it below
  type queue is array (0 to 31) of std_logic_vector(20 downto 0);
  signal write_q : queue;

  signal back_q : integer range 0 to 31 := 0;
  signal front_q : integer range 0 to 31 := 0;

  signal write_phase : std_logic := '0'; -- For two-phase SRAM writes

  -- Where in the 1024-row space are we looking at? 10 bits for unsigned integer up
  -- to 1024, extra bit for sub-pixel scrolling accuracy since pixels are doubled
  signal vert_offset : unsigned(10 downto 0) := (others => '0');

  -- Tile matrix: 10 tiles across, 6 rows

  -- tilemat has 50 positions for 5 rows of 10 tiles
  type tilematrix_t is array(0 to 49) of unsigned(20 downto 0);
  signal tilemat : tilematrix_t;

  -- TODO: OPTimize the way we're storing animation state cuz it's being stored as
  -- a global for all tiles

  -- Each entry has these bits:
  -- 6 downto 0: tile ID, the thing we use to index into the tiles array
  -- 13 downto 7: tile ID of the Tile this Tile will be morphing into
  -- 17 downto 14: animation state
  -- 19 downto 18: which corner is the tile in
  -- 20: should the tile be tinted

  -- Does the row that we're currently scanning out have tiles on it?
  -- Should only be set when incanvas = '0', i.e., when we're on the left half.
  -- NOTE: not all rows have tiles
  signal row_of_tiles : std_logic := '0';

  -- curenttile: Which of the 64 tiles are we currently in the process of setting up?
  signal currenttile : unsigned(6 downto 0) := (others => '0');

  --tile_ptr: Where in the tile is our next set of 4 pixels (4 bits each) being written
  -- to? i.e. which set of 4 pixels are we writing to next?
  signal tile_ptr : unsigned(7 downto 0) := (others => '0');

  -- Which index in the tile matrix are we going to be updating?
  signal tile_to_change : integer range 0 to 49 := 0;

  --Signals used for RAM:
  signal weRamTiles: std_logic := '0'; --write enable for ram block of array tiles.
  --address for ram block of array tiles:
  signal addRamTiles: unsigned(14 downto 0) := (others => '0');
  --data of Tile information being sent to be saved in Ram block:
  signal diRamTiles: unsigned(15 downto 0) := (others => '0');
  --data of Tile information being read from Ram block.
  signal doRamTiles: unsigned(15 downto 0) := (others => '0');

  signal tint: std_logic := '0';
  signal animationStateHolder: integer;

  signal resetCounter: std_logic :='0';
  signal countervalue: integer;

begin

  comp: TilesRam PORT MAP (
    clk => clk,
    we => weRamTiles,
    a => addRamTiles,
    di => diRamTiles,
    do => doRamTiles
  );

  comp1: counter PORT MAP(
    clk => clk,
    reset => resetCounter,
    q => countervalue
  );

	LEDR(7 downto 0) <= GPIO_0(7 downto 0);

  process (clk)
  begin
    if rising_edge(clk) then
      clk25 <= not clk25;
    end if;
  end process;

  SRAM_UB_N <= '0'; -- Upper Byte
  SRAM_LB_N <= '0'; -- Lower byte
  SRAM_CE_N <= '0'; -- Chip enable
  SRAM_OE_N <= '0'; -- Read enable---keep it enabled even during writes

  process (clk)

  variable q_address : unsigned(4 downto 0) := (others => '0');
  variable q_writedata : unsigned(15 downto 0) := (others => '0');

  variable x : unsigned(9 downto 0) := (others => '0');
  variable y : unsigned(9 downto 0) := (others => '0');
  variable compensated_x : unsigned(9 downto 0) := (others => '0');

  -- At which vertical set of 32 pixels are we?
  variable row : unsigned(3 downto 0) := (others => '0');

  variable tilemat_index : integer range 0 to 49 := 0;
  variable holder : integer;

  variable animationSize: std_logic := '0';-- if 0 then small if 1 then large
  variable animationState: unsigned(2 downto 0);

  begin
    if rising_edge(clk) then
      if reset = '1' then
        write_pointer <= (others => '0');
      else
        x := Hcount - HSYNC - HBACK_PORCH;
        y := Vcount - VSYNC - VBACK_PORCH;

        if chipselect = '1' then
          if read = '1' then
            readdata(9 downto 0) <= vert_offset(10 downto 1);
          elsif write = '1' then
            write_q(front_q)(20 downto 16) <= std_logic_vector(address);
            write_q(front_q)(15 downto 0) <= std_logic_vector(writedata);
            front_q <= front_q + 1;
          end if;
        end if;

        if incanvas = '1' then
          SRAM_WE_N <= '1';
          SRAM_DQ <= (others => 'Z');
          SRAM_ADDR(17 downto 8) <= std_logic_vector(read_pointer(19 downto 10));
          SRAM_ADDR(7 downto 0) <= std_logic_vector(read_pointer(8 downto 1));

        elsif row_of_tiles = '1' and write_phase = '0' then
          row := y(8 downto 5);
          case row is
            when x"2" => tilemat_index := 0;
            when x"3" => tilemat_index := 10;
            when x"5" => tilemat_index := 20;
            when x"8" => tilemat_index := 30;
            when x"A" => tilemat_index := 40;
            when others => tilemat_index := 0;
          end case;

          compensated_x := x + 1; --set block ram address ahead to read properly 
          tilemat_index := tilemat_index + TO_INTEGER(compensated_x(8 downto 5));

          -- ** ANIMATION **

          tint <= tilemat(tilemat_index)(20);

          animationState := tilemat(tilemat_index)(16 downto 14);

          animationSize := tilemat(tilemat_index)(17);

          -- If animation state is 0, clear
          if tilemat(tilemat_index)(16 downto 14) = "000" then
            addRamTiles <= (others => '0');

	      -- if animation state is 1, show tile
          elsif tilemat(tilemat_index)(16 downto 14)= "001" then
			--7 bit address of tile ID:
            addRamTiles(14 downto 8) <= tilemat(tilemat_index)(6 downto 0);
            addRamTiles(7 downto 3) <= y(4 downto 0);
            addRamTiles(2 downto 0) <= compensated_x(4 downto 2);

          -- if animation state is 2 only need 1 value from NIOS, rest can be done in HW
          elsif tilemat(tilemat_index)(16 downto 14)= "010" then
            addRamTiles(2 downto 0) <= compensated_x(4 downto 2);

            if animationSize = '0' then -------------If tiles are small

              if animationStateHolder < 10 then

                if y(4 downto 0) < "10000" then
                  holder := animation(animationStateHolder*16+to_integer(y(4 downto 0)));

                  if holder = 32 then
                     -- display next tile:
                    addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);
                    -- non-squeezed tile next:
                    addRamTiles(7 downto 3) <= y(4 downto 0);

                  else
                    -- display current tile:
                    addRamTiles(14 downto 8) <= tilemat(tilemat_index)(6 downto 0);
                    -- squeezed tile current:
                    addRamTiles(7 downto 3) <= to_unsigned(holder, 5);
                  end if;

                else
                  -- 7 bit address of tile ID:
                  addRamTiles(14 downto 8) <= tilemat(tilemat_index)(6 downto 0);
                  addRamTiles(7 downto 3) <= y(4 downto 0);

                end if;

              elsif animationStateHolder = 10 then
                addRamTiles(7 downto 3) <= y(4 downto 0);

                if y(4 downto 0) < "10000" then
                  addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);

                else
                  addRamTiles(14 downto 8) <= tilemat(tilemat_index)(6  downto 0);
                end if;

              elsif animationStateHolder<14 and animationStateHolder>10 then

                if y(4 downto 0) < "10000" then
                  addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);
                  addRamTiles(7 downto 3) <= y(4 downto 0);

                else
                  holder := animation((animationStateHolder-1)*16+to_integer(y(3 downto 0)));

                  if (holder)=32 then
                    addRamTiles(14 downto 8) <= tilemat(tilemat_index)(6 downto 0);
                    addRamTiles(7 downto 3) <= y(4 downto 0);

                  else
                    addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);
                    addRamTiles(7 downto 3) <= to_unsigned(holder, 5);

                  end if;

                end if;

              else
				-- 7 bit address of tile ID:
                addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);
                addRamTiles(7 downto 3) <= y(4 downto 0);
              end if;

            else--------------------------------If the tiles are large

              if animationStateHolder<10 then

                if tilemat_index > -1 and tilemat_index < 10 then

                  holder := animationBIG(animationStateHolder*32+to_integer(y(4 downto 0)));

                  if holder = 32 then
                    -- display next tile:
                    addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);
				    -- non-squeezed tile next:
                    addRamTiles(7 downto 3) <= y(4 downto 0);

                  else
                    -- display current tile:
                    addRamTiles(14 downto 8) <= tilemat(tilemat_index)(6 downto 0);
		            -- squeezed tile current
                    addRamTiles(7 downto 3) <= to_unsigned(holder,5);
                  end if;

                elsif ((tilemat_index >(9))and(tilemat_index<20))then
                  --7 bit address of tile ID:
                  addRamTiles(14 downto 8) <= tilemat(tilemat_index)(6 downto 0);
                  addRamTiles(7 downto 3) <= y(4 downto 0);
                end if;

              elsif animationStateHolder = 10 then
				addRamTiles(7 downto 3) <= y(4 downto 0);
				
                if ((tilemat_index >(-1))and(tilemat_index<10)) then
                  addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);

                elsif ((tilemat_index >(9))and(tilemat_index<20))then
                  addRamTiles(14 downto 8) <= tilemat(tilemat_index)(6 downto 0);
                end if;

              elsif animationStateHolder<14 and animationStateHolder>10 then

                if ((tilemat_index >(-1))and(tilemat_index<10)) then
                  -- 7 bit address of tile ID:
                  addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);
                  addRamTiles(7 downto 3) <= y(4 downto 0);

                elsif ((tilemat_index >(9))and(tilemat_index<20))then
                  holder := animationBIG((animationStateHolder-1)*32+to_integer(y(4 downto 0)));

                  if (holder)=32 then
                    -- 7 bit address of tile ID:
                    addRamTiles(14 downto 8) <= tilemat(tilemat_index)(6 downto 0);
                    addRamTiles(7 downto 3) <= y(4 downto 0);

                  else
                    -- 7 bit address of tile ID:
                    addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);
                    addRamTiles(7 downto 3) <= to_unsigned(holder,5);
                  end if;

                end if;

              else
                -- 7 bit address of tile ID:
                addRamTiles(14 downto 8) <= tilemat(tilemat_index)(13 downto 7);
                addRamTiles(7 downto 3) <= y(4 downto 0);
              end if;

            end if;

            if y(3 downto 0) = "1111"  and countervalue = 1428570 then
              animationStateHolder <= animationStateHolder + 1;
              resetCounter <= '1';

            else
              resetCounter <= '0';
            end if;

          end if;

          -- ** END_ANIMATION **

        elsif chipselect = '0' then
          if (front_q /= back_q) and (write_phase = '0') then
            q_address := unsigned(write_q(back_q)(20 downto 16));
            q_writedata := unsigned(write_q(back_q)(15 downto 0));

            if q_address = OP_WRITE then
              SRAM_WE_N <= '0';
              SRAM_ADDR <= std_logic_vector(write_pointer);
              SRAM_DQ <= std_logic_vector(q_writedata);
              write_pointer <= write_pointer + 1;

            elsif q_address = OP_SET_VERT_OFFSET then
              vert_offset(10 downto 1) <= q_writedata(9 downto 0);

            elsif q_address = OP_SEEK_ROW then
              write_pointer(17 downto 8) <= q_writedata(9 downto 0);
              write_pointer(7 downto 0) <= (others => '0');

            elsif q_address = OP_TILE_SET_CURRENT then
              currenttile <= q_writedata(6 downto 0);
              tile_ptr <= (others => '0');

                 elsif q_address = OP_TILE_SEEK then
                   tile_ptr <= q_writedata(7 downto 0);

                 elsif q_address = OP_TILE_WRITE then
                   weRamTiles <= '1';
                   addRamTiles(7 downto 0) <= tile_ptr;
                   addRamTiles(14 downto 8) <= currenttile;
                   diRamTiles <= q_writedata;
                   tile_ptr <= tile_ptr + 1;

                  elsif q_address = OP_TILE_TO_CHANGE then
                    tile_to_change <= TO_INTEGER(q_writedata(5 downto 0));
					LEDG(1)<='0';

				elsif q_address = OP_TINT then
                    tilemat(tile_to_change)(20) <= q_writedata(0);

				elsif q_address = OP_TILE_TRANSFORMER_CHANGE then
                    tilemat(tile_to_change)(13 downto 7) <= q_writedata(6 downto 0);
					animationStateHolder<=0;
					LEDG(1)<='1';

                  elsif q_address = OP_TILE_CHANGE_TO then
                    tilemat(tile_to_change)(6 downto 0) <= q_writedata(6 downto 0);

				elsif q_address = OP_TILE_SET_ANIM_STATE then
                    tilemat(tile_to_change)(17 downto 14) <= q_writedata(3 downto 0);
                end if;

                write_phase <= not write_phase;

            elsif write_phase = '1' then -- End the write
                SRAM_WE_N <= '1';
                SRAM_DQ <= (others => 'Z');
                weRamTiles <= '0';
                back_q <= back_q + 1;
                write_phase <= not write_phase;
            end if;
        end if;

        if x = 0 and y = 481 and clk25 = '1' then -- Once outside the screen
            vert_offset <= vert_offset + 1;
        end if;

      end if;
    end if;
  end process;


  -- Horizontal and vertical counters

  HCounter : process (clk25)
  begin
    if rising_edge(clk25) then
      if reset = '1' then
        Hcount <= (others => '0');
      elsif EndOfLine = '1' then
        Hcount <= (others => '0');
      else
        Hcount <= Hcount + 1;
      end if;
    end if;
  end process HCounter;

  EndOfLine <= '1' when Hcount = HTOTAL - 1 else '0';

  VCounter: process (clk25)
  begin
    if rising_edge(clk25) then
      if reset = '1' then
        Vcount <= (others => '0');
      elsif EndOfLine = '1' then
        if EndOfField = '1' then
          Vcount <= (others => '0');
        else
          Vcount <= Vcount + 1;
        end if;
      end if;
    end if;
  end process VCounter;

  EndOfField <= '1' when Vcount = VTOTAL - 1 else '0';

  -- State machines to generate HSYNC, VSYNC, HBLANK, and VBLANK

  HSyncGen : process (clk25)
  begin
    if rising_edge(clk25) then
      if reset = '1' or EndOfLine = '1' then
        vga_hsync <= '1';
      elsif Hcount = HSYNC - 1 then
        vga_hsync <= '0';
      end if;
    end if;
  end process HSyncGen;

  HBlankGen : process (clk25)
  begin
    if rising_edge(clk25) then
      if reset = '1' then
        vga_hblank <= '1';
      elsif Hcount = HSYNC + HBACK_PORCH then
        vga_hblank <= '0';
      elsif Hcount = HSYNC + HBACK_PORCH + HACTIVE then
        vga_hblank <= '1';
      end if;
    end if;
  end process HBlankGen;

  VSyncGen : process (clk25)
  begin
    if rising_edge(clk25) then
      if reset = '1' then
        vga_vsync <= '1';
      elsif EndOfLine ='1' then
        if EndOfField = '1' then
          vga_vsync <= '1';
        elsif Vcount = VSYNC - 1 then
          vga_vsync <= '0';
        end if;
      end if;
    end if;
  end process VSyncGen;

  VBlankGen : process (clk25)
  begin
    if rising_edge(clk25) then
      if reset = '1' then
        vga_vblank <= '1';
      elsif EndOfLine = '1' then
        if Vcount = VSYNC + VBACK_PORCH - 1 then
          vga_vblank <= '0';
        elsif Vcount = VSYNC + VBACK_PORCH + VACTIVE - 1 then
          vga_vblank <= '1';
        end if;
      end if;
    end if;
  end process VBlankGen;

  -- detects if we're inside the canvas, which is the right half
  -- Does other bookkeeping as well
  IfInCanvas : process (clk25)
    variable x : unsigned(9 downto 0) := (others => '0');
    variable y : unsigned(9 downto 0) := (others => '0');

    -- Which vertical set of 32 pixels are we at?
    variable row : unsigned(3 downto 0) := (others => '0');
  begin
    if rising_edge(clk25) then
      x := Hcount - HSYNC - HBACK_PORCH;
      y := Vcount - VSYNC - VBACK_PORCH;

      if y >= 0 and y < 480 and x >= 320 and x < 640 then
        incanvas <= '1';
        row_of_tiles <= '0';
        read_pointer <= read_pointer + 1;
      else
        incanvas <= '0';
        row := y(8 downto 5); -- y divided by 32
        if (row = 2 or row = 3 or row = 5 or row = 8 or row = 10)
          and (x < 320 and x >= 0) then
          -- We're in an area potentially occluded by tiles
          row_of_tiles <= '1';
        else
          row_of_tiles <= '0';
        end if;
      end if;

      if x = 23 then
        -- Random column outside canvas designated to setting
        -- up the SRAM address in preparation for reading out
        -- image data.
        read_pointer(19 downto 9) <= vert_offset + y;
        read_pointer(8 downto 0) <= "111111111";
      end if;
    end if;
  end process IfInCanvas;

  -- Registered video signals going to the video DAC
  VideoOut: process (clk25, reset)

  -- Coordinate of the current pixel position on the raster
  variable x : unsigned(9 downto 0) := (others => '0');
  variable y : unsigned(9 downto 0) := (others => '0');

  -- The x and y coordinate on the background gradient
  -- which is just the coordinate plus the brightness offset
  -- from GPIO_0(7 downto 0). x gives the redness and y gives
  -- the blueness of the background---these variable help
  -- simplify the blending code
  variable bg_x : unsigned(9 downto 0) := (others => '0');
  variable bg_y : unsigned(9 downto 0) := (others => '0');

  -- 4-bit alpha of the pixel of the tile we're currently at
  -- variable used to simplify code for blending
  variable alpha : unsigned(3 downto 0) := (others => '0');
  
  variable vga_red : unsigned(4 downto 0) := (others => '0');
  variable vga_green : unsigned(5 downto 0) := (others => '0');
  variable vga_blue : unsigned(4 downto 0) := (others => '0');

  -- The foreground color of the tiles---this isn't constant
  -- because it changes with the y-coordinate to simulate
  -- shading
  variable C511 : unsigned(8 downto 0) := (others => '1');

  constant C15   : unsigned(3 downto 0) := (others => '1');
  variable get_y : unsigned(3 downto 0) := (others => '0');

  variable temp13 : unsigned(12 downto 0) := (others => '0');

  begin
    if reset = '1' then
      VGA_R <= "0000000000";
      VGA_G <= "0000000000";
      VGA_B <= "0000000000";
    elsif rising_edge(clk25) then
      x := Hcount - HSYNC - HBACK_PORCH;
      y := Vcount - VSYNC - VBACK_PORCH;

      bg_x := x + unsigned(GPIO_0(7 downto 0));
      bg_y := y + unsigned(GPIO_0(7 downto 0));
	  get_y := unsigned(465 - y)(3 downto 0);

      if(y > 128) then
        -- Foreground color shading for small tiles
        C511(7 downto 3) := "11111" - y(4 downto 0);
      else
        -- Foreground color shading for large tiles
        C511(7 downto 1) := "1111111" - y(6 downto 0) + "011000";
      end if;

      if (x >= 0 and x < 640) and (y >= 0 and y < 480) then

        -- if we're in the right hand side--the part with images:
        if incanvas = '1' and SRAM_DQ /= "0000011111111111" then
          -- if it's not a transparent tile, then take RGB info from
          -- SRAM output---SRAM address has been set in another process
          
--		  if (y < 16) or (y > 464) then
--		
--			vga_red(4 downto 0) := unsigned(SRAM_DQ(15 downto 11));
--            vga_green(5 downto 0) := unsigned(SRAM_DQ(10 downto 5));
--            vga_blue(4 downto 0) := unsigned(SRAM_DQ(4 downto 0));
--			
--			if y < 16 then
--				
--				temp13(12 downto 4) := vga_red * y(3 downto 0);
--				temp13 := temp13 + (bg_x(9 downto 1) * (C15 - y(3 downto 0)));
--				VGA_R <= std_logic_vector(temp13(12 downto 3));
--				
--				VGA_G <= std_logic_vector(vga_green * y(3 downto 0));
--				
--				temp13(12 downto 4) := vga_blue * y(3 downto 0);
--				temp13 := temp13 + (bg_y(9 downto 1) * (C15 - y(3 downto 0)));
--				VGA_B <= std_logic_vector(temp13(12 downto 3));
--			
--			else
--				
--				temp13(12 downto 4) := vga_red * get_y;
--				temp13 := temp13 + (bg_x(9 downto 1) * (C15 - get_y));
--				VGA_R <= std_logic_vector(temp13(12 downto 3));
--				
--				VGA_G <= std_logic_vector(vga_green * get_y);
--				
--				temp13(12 downto 4) := vga_blue * get_y;
--				temp13 := temp13 + (bg_y(9 downto 1) * (C15 - get_y));
--				VGA_B <= std_logic_vector(temp13(12 downto 3));
--			end if;
--				
--		  else
		
		    VGA_R(9 downto 5) <= SRAM_DQ(15 downto 11);
            VGA_G(9 downto 4) <= SRAM_DQ(10 downto 5);
            VGA_B(9 downto 5) <= SRAM_DQ(4 downto 0);	

--		  end if;
		
        elsif row_of_tiles = '1' then
          -- if we're in an area that may be occluded by tiles
          -- TODO: can you rewrite this as with input select output?
          case x(1 downto 0) is
            when "00" => alpha := doRamTiles(15 downto 12);
            when "01" => alpha := doRamTiles(11 downto 8);
            when "10" => alpha := doRamTiles(7 downto 4);
            when "11" => alpha := doRamTiles(3 downto 0);
            when others => alpha := (others => '0');
          end case;

          if alpha = x"0" then
            -- Show the background---no occlusion
            VGA_R <= std_logic_vector(bg_x);
            VGA_B <= std_logic_vector(bg_y);
            VGA_G <= (others => '0');
          else
            if tint = '1' then
              temp13 := ("101011100" * alpha) + (bg_x(9 downto 1) * (C15 - alpha));
              VGA_R <= std_logic_vector(temp13(12 downto 3));

              temp13 := ("011111100" * alpha) + bg_y(9 downto 1) * (C15 - alpha);
              VGA_B <= std_logic_vector(temp13(12 downto 3));
            else
              temp13 := (C511 * alpha) + (bg_x(9 downto 1) * (C15 - alpha));
              VGA_R <= std_logic_vector(temp13(12 downto 3));

              temp13 := (C511 * alpha) + (bg_y(9 downto 1) * (C15 - alpha));
              VGA_B <= std_logic_vector(temp13(12 downto 3));
            end if;

            temp13 := (C511 * alpha);
            VGA_G <= std_logic_vector(temp13(12 downto 3));
          end if;
        else
          -- Show the background -- no occlusion
          VGA_R <= std_logic_vector(bg_x);
          VGA_B <= std_logic_vector(bg_y);
          VGA_G <= (others => '0');
        end if;

      elsif vga_hblank = '0' and vga_vblank ='0' then
        VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";
      else
        VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";
      end if;
    end if;
  end process VideoOut;

  VGA_CLK <= clk25;
  VGA_HS <= not vga_hsync;
  VGA_VS <= not vga_vsync;
  VGA_SYNC <= '0';
  VGA_BLANK <= not (vga_hsync or vga_vsync);

end rtl;


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity TilesRam is
  port (
    clk : in std_logic;
    we : in std_logic;
    a : in unsigned(14 downto 0); --need 15 bits to access all units in array
    di : in unsigned(15 downto 0);
    do : out unsigned(15 downto 0)
  );
end TilesRam;

architecture rtl of TilesRam is
  type ram_type is array (0 to 18431) of unsigned(15 downto 0);
  -- 4 pixels are stored per row for convenience, 256*71, 71 tiles, with 16*16 pixels
  -- ...or did we change it to 72 tiles?
  signal RAM : ram_type;
  signal read_a : unsigned(14 downto 0);
begin
  process (clk)
  begin
    if rising_edge(clk) then
      if we = '1' then
        RAM(to_integer(a)) <= di;
      end if;
      read_a <= a;
    end if;
  end process;
  do <= RAM(to_integer(read_a));
end rtl;


-- TODO: the code below is redundant. we already have
-- a counter in the timer module only with a different
-- timeout value.

-- TODO: we can update animation states without a timer
-- by leveraging the VGA pixel timer, which runs at about
-- the same frequency by chance

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity counter is
  port(
    Clk, Reset : in std_logic;
    Q : out integer
  );
end counter;

architecture imp of counter is
  signal count : integer range 0 to 1428571 := 0;
begin
  process (Clk,Reset,count)
  begin
    if Reset = '1' then
      Q <= 0;
    else
      Q <= count;
    end if;

    if rising_edge(Clk) then
      if Reset = '1' then
        count <= 0;
      else
        if count < 1428570 then
          count <= count + 1;
        end if;
      end if;
    end if;
  end process;
end imp;
